@@ -79,11 +79,15 @@ oil_test_new (OilFunctionClass *klass)
   test->klass = klass;
   test->proto = proto;
   test->impl = klass->reference_impl;
+  test->tolerance = 0.0;
 
   for (i=0;i<proto->n_params;i++){
     if (proto->params[i].parameter_type == OIL_ARG_UNKNOWN) {
       return NULL;
     }
+    if (oil_type_is_floating_point(proto->params[i].type)) {
+      test->tolerance = 0.001;
+    }
     memcpy (&test->params[proto->params[i].parameter_type], &proto->params[i],
         sizeof(OilParameter));
   }
@@ -91,6 +95,8 @@ oil_test_new (OilFunctionClass *klass)
     test->params[i].src_data = NULL;
     test->params[i].ref_data = NULL;
     test->params[i].test_data = NULL;
+    test->params[i].test_header = OIL_TEST_HEADER;
+    test->params[i].test_footer = OIL_TEST_FOOTER;
   }
 
   test->iterations = 10;
@@ -156,6 +162,34 @@ oil_test_set_iterations (OilTest *test, int iterations)
   test->iterations = iterations;
 }
 
+/**
+ * oil_test_set_test_header:
+ * @test: the OilTest
+ * @p: the OilParameter to change the header for
+ * @test_header: the number of bytes of guard header
+ *
+ * Sets the number of bytes of guard header for @p to @test_header.
+ */
+void
+oil_test_set_test_header (OilTest *test, OilParameter *p, int test_header)
+{
+  p->test_header = test_header;
+}
+
+/**
+ * oil_test_set_test_footer:
+ * @test: the OilTest
+ * @p: the OilParameter to change the footer for
+ * @test_footer: the number of bytes of guard footer
+ *
+ * Sets the number of bytes of guard footer for @p to @test_footer.
+ */
+void
+oil_test_set_test_footer (OilTest *test, OilParameter *p, int test_footer)
+{
+  p->test_footer = test_footer;
+}
+
 /**
  * oil_test_init:
  * @test: the OilTest
@@ -206,13 +240,13 @@ oil_test_check_function (void * priv)
     if (p->is_pointer) {
       pointer_mask |= 1;
       if (p->direction == 's') {
-        args[i] = (unsigned long)p->src_data + OIL_TEST_HEADER;
+        args[i] = (unsigned long)p->src_data + p->test_header;
       } else if (p->direction == 'i') {
         memcpy (p->test_data, p->src_data, p->size);
-        args[i] = (unsigned long)p->test_data + OIL_TEST_HEADER;
+        args[i] = (unsigned long)p->test_data + p->test_header;
       } else if (p->direction == 'd') {
         memset (p->test_data, p->guard, p->size);
-        args[i] = (unsigned long)p->test_data + OIL_TEST_HEADER;
+        args[i] = (unsigned long)p->test_data + p->test_header;
       } else {
         OIL_ERROR ("not reached");
       }
@@ -237,8 +271,8 @@ oil_test_check_function (void * priv)
         pointer_mask, &test->prof);
   }
 
-  oil_profile_get_ave_std (&test->prof, &test->impl->profile_ave,
-      &test->impl->profile_std);
+  oil_profile_get_ave_std (&test->prof, &test->profile_ave,
+      &test->profile_std);
 }
 
 /**
@@ -323,8 +357,8 @@ oil_test_check_impl (OilTest *test, OilFunctionImpl *impl)
   ret = oil_cpu_fault_check_try (oil_test_check_function, test);
   if (!ret) {
     OIL_ERROR ("illegal instruction in %s", test->impl->name);
-    test->impl->profile_ave = 0;
-    test->impl->profile_std = 0;
+    test->profile_ave = 0;
+    test->profile_std = 0;
 
     return 0;
   }
@@ -336,16 +370,17 @@ oil_test_check_impl (OilTest *test, OilFunctionImpl *impl)
 
     if (p->is_pointer) {
       if (p->direction == 'i' || p->direction == 'd') {
-        x += check_array (p->test_data, p->ref_data, p->type, p->pre_n,
-            p->stride, p->post_n);
+        x += check_array (p->test_data + p->test_header,
+            p->ref_data + p->test_header, p->type, p->pre_n, p->stride,
+            p->post_n);
         n += p->pre_n * p->post_n;
-        if (!check_guard (p->test_data, OIL_TEST_HEADER, p->guard)) {
+        if (!check_guard (p->test_data, p->test_header, p->guard)) {
           fail = 1;
           OIL_ERROR("function %s wrote before area for parameter %s",
               test->impl->name, p->parameter_name);
         }
-        if (!check_guard ((uint8_t *)p->test_data + p->size - OIL_TEST_FOOTER,
-              OIL_TEST_FOOTER, p->guard)) {
+        if (!check_guard ((uint8_t *)p->test_data + p->size - p->test_footer,
+              p->test_footer, p->guard)) {
           fail = 1;
           OIL_ERROR("function %s wrote after area for parameter %s",
               test->impl->name, p->parameter_name);
@@ -363,9 +398,9 @@ oil_test_check_impl (OilTest *test, OilFunctionImpl *impl)
   test->sum_abs_diff = x;
   test->n_points = n;
 
-  if (x > n || fail) {
-    OIL_ERROR ("function %s in class %s failed check (%g > %d) outside=%d",
-        test->impl->name, test->klass->name, x, n, fail);
+  if (x > test->tolerance * n || fail) {
+    OIL_ERROR ("function %s in class %s failed check (%g > %g) || (outside=%d)",
+        test->impl->name, test->klass->name, x, test->tolerance * n, fail);
     return 0;
   }
 
@@ -455,7 +490,7 @@ init_parameter (OilTest *test, OilParameter *p, OilParameter *ps)
     }
   }
 
-  p->size = p->stride * p->post_n + OIL_TEST_HEADER + OIL_TEST_FOOTER;
+  p->size = p->stride * p->post_n + p->test_header + p->test_footer;
   p->guard = oil_rand_u8();
 
   if (p->direction == 'i' || p->direction == 's') {
@@ -464,7 +499,7 @@ init_parameter (OilTest *test, OilParameter *p, OilParameter *ps)
     OIL_DEBUG("allocating %d bytes for src_data for %s", p->size, p->parameter_name);
     p->src_data = malloc (p->size);
     memset (p->src_data, p->guard, p->size);
-    fill_array (p->src_data + OIL_TEST_HEADER, p->type, p->pre_n, p->stride, p->post_n);
+    fill_array (p->src_data + p->test_header, p->type, p->pre_n, p->stride, p->post_n);
   }
 
   if (p->direction == 'i' || p->direction == 'd') {
@@ -508,47 +543,43 @@ static void
 fill_array (void *data, OilType type, int pre_n, int stride, int post_n)
 {
   int i;
-  int j;
-  int s2 = oil_type_sizeof (type);
 
 #define FILL(type,func) do {\
   for(i=0;i<post_n;i++){ \
-    for(j=0;j<pre_n;j++){ \
-      OIL_GET(data, i*stride + j*s2, type) = func ; \
-    } \
+    func (OIL_OFFSET(data, i*stride), pre_n); \
   } \
 }while(0)
 
   switch (type) {
     case OIL_TYPE_s8p:
-      FILL(int8_t,oil_rand_s8());
+      FILL(int8_t,oil_random_s8);
       break;
     case OIL_TYPE_u8p:
-      FILL(uint8_t,oil_rand_u8());
+      FILL(uint8_t,oil_random_u8);
       break;
     case OIL_TYPE_s16p:
-      FILL(int16_t,oil_rand_s16());
+      FILL(int16_t,oil_random_s16);
       break;
     case OIL_TYPE_u16p:
-      FILL(uint16_t,oil_rand_u16());
+      FILL(uint16_t,oil_random_u16);
       break;
     case OIL_TYPE_s32p:
-      FILL(int32_t,oil_rand_s32());
+      FILL(int32_t,oil_random_s32);
       break;
     case OIL_TYPE_u32p:
-      FILL(uint32_t,oil_rand_u32());
+      FILL(uint32_t,oil_random_u32);
       break;
     case OIL_TYPE_s64p:
-      FILL(int64_t,oil_rand_s64());
+      FILL(int64_t,oil_random_s64);
       break;
     case OIL_TYPE_u64p:
-      FILL(uint64_t,oil_rand_u64());
+      FILL(uint64_t,oil_random_u64);
       break;
     case OIL_TYPE_f32p:
-      FILL(float,oil_rand_f32_0_1());
+      FILL(float,oil_random_f32);
       break;
     case OIL_TYPE_f64p:
-      FILL(double,oil_rand_f64_0_1());
+      FILL(double,oil_random_f64);
       break;
     default:
       OIL_ERROR ("should not be reached (type == %d)", type);
@@ -565,6 +596,11 @@ check_array (void *data, void *ref, OilType type, int pre_n, int stride, int pos
   int s2 = oil_type_sizeof (type);
   double x = 0;
 
+#if 0
+  OIL_ERROR ("check array pre_n=%d stride=%d post_n=%d",
+      pre_n, stride, post_n);
+#endif
+
 #define CHECK(type) do {\
   for(i=0;i<post_n;i++){ \
     for(j=0;j<pre_n;j++){ \
@@ -636,3 +672,19 @@ check_holes (void *data, OilType type, int pre_n, int stride, int post_n,
   return 1;
 }
 
+void *
+oil_test_get_source_data (OilTest *test, OilArgType arg_type)
+{
+  uint8_t *ptr;
+ 
+  ptr = test->params[arg_type].src_data;
+  ptr += test->params[arg_type].test_header;
+
+  return ptr;
+}
+
+int
+oil_test_get_value (OilTest *test, OilArgType arg_type)
+{
+  return test->params[arg_type].value;
+}
